/*
	copyright Super(xia.tony@gmail.com)
*/

#include "dl.h"
#include "../std.h"
#include "../master.h"
#include "../interpret.h"

#include <stdio.h>

typedef void* DL_HANDLE;

typedef struct list_s
{
	DL_HANDLE h;
	struct list_s *next;
	struct list_s *pre;
}list_t;

list_t *head;
static int initialized = 0;

void init()
{
	initialized = 1;
	head = NULL;
}

DL_HANDLE open_dl(const char *fname)
{
	DL_HANDLE f;
	char *err_msg;
	list_t *new_node;

	if(!initialized) init();

	f = dlopen(fname,RTLD_LAZY);
	err_msg = (char *)dlerror();
	if(err_msg != NULL)
	{
		error(err_msg);
		return 0;
	}
	
	new_node = (list_t*)malloc(sizeof(list_t));
	new_node->h = f;
	new_node->next = head;
	new_node->pre = 0;
	if(head != NULL) head->pre = new_node;
	head = new_node;
	return f;
}

int remove_dl(DL_HANDLE h)
{
	char *err_msg;
	list_t *node;

	dlclose(h);
	if((err_msg = (char *)dlerror()) != NULL)
	{
		error(err_msg);
		return 0;
	}

	node = head;
	while(node != NULL)
	{
		if(node->h == h)
		{
			if(node->pre) node->pre->next = node->next;
			if(node->next) node->next->pre = node->pre;
			if(node == head) head = node->next;
			free(node);
			node = 0;
			break;
		}
	}

	return 1;
}

#ifdef F_DL_CALL

void f_dl_call()
{
	svalue_t *v;
	char *func_name;
	void *f;
	int arg_num, pushed, offset;
	list_t *node;
	
	if(!st_num_arg) error("At least should have 1 argument.\n");

	v = sp - st_num_arg + 1;
	if(v->type != T_STRING)
		error("First argument should be string.\n");

	func_name = v->u.string;
	f = NULL;
	node = head;
	while(f == NULL && node != NULL)
	{
		f = dlsym(node->h, func_name);
		node = node->next;
	}
	if(f == NULL)
		error("No such function in loaded dls.\n");

	arg_num = st_num_arg - 1;
	pushed = 0;
	while(arg_num--)
	{
		svalue_t *sv = sp - arg_num;
		switch(sv->type)
		{
			case T_STRING :	__asm__("push %0"::"r"(sv->u.string)); pushed ++; break;
			case T_NUMBER :	__asm__("push %0"::"r"(sv->u.number)); pushed ++; break;
			case T_REAL :	__asm__("push %0"::"r"(sv->u.real)); pushed ++; break;
			default : error("Do not support this type of argument now.\n");
		}
	}
	
	offset = pushed * 0x04;
	__asm__("call * %0"::"r"(f));
	__asm__("add %%esp, %0"::"r"(offset));
	pop_n_elems(st_num_arg);
}

#endif

#ifdef F_DL_OPEN
void f_dl_open()
{
	char buf[10];
	DL_HANDLE res = open_dl(sp->u.string);
	free_svalue(sp, "f_dl_open");
	sprintf(buf, "%ld", (long)res);
	sp->type = T_STRING;
	sp->u.string = make_shared_string(buf);
	sp->subtype = STRING_SHARED;
}
#endif

#ifdef F_DL_CLOSE

DL_HANDLE get_address_from_str(char *str)
{
	long addr;
	addr = atoi(str);
	return (DL_HANDLE)addr;
}

void f_dl_close()
{
	DL_HANDLE f = get_address_from_str(sp->u.string);
	free_svalue(sp, "f_dl_close");
	sp->type = T_NUMBER;
	sp->u.number = remove_dl(f);
}
#endif
